from typing import Optional
import torch

from torchio.data.subject import Subject
from torchio.data.image import ScalarImage
from torchio.transforms.intensity_transform import IntensityTransform
from torchio.transforms.preprocessing.intensity.clamp import Clamp


class Threshold(IntensityTransform):
    """Threshold intensity values into a range :math:`[a, b]`.

    Args:
        out_min: Minimum value :math:`a` of the output image. If ``None``, the
            minimum of the image is used.
        out_max: Maximum value :math:`b` of the output image. If ``None``, the
            maximum of the image is used.

    Example:
        >>> import torchio as tio
        >>> ct = tio.datasets.Slicer('CTChest').CT_chest
        >>> HOUNSFIELD_AIR, HOUNSFIELD_BONE = -1000, 1000
        >>> threshold = tio.Threshold(out_min=HOUNSFIELD_AIR, out_max=HOUNSFIELD_BONE)
        >>> ct_thresholded = threshold(ct)

    .. plot::

        import torchio as tio
        subject = tio.datasets.Slicer('CTChest')
        ct = subject.CT_chest
        HOUNSFIELD_AIR, HOUNSFIELD_BONE = -1000, 1000
        threshold = tio.Threshold(out_min=HOUNSFIELD_AIR, out_max=HOUNSFIELD_BONE)
        ct_thresholded = threshold(ct)
        subject.add_image(ct_thresholded, 'Thresholded')
        subject.plot()

    """

    def __init__(
        self, out_min: Optional[float] = None, out_max: Optional[float] = None, **kwargs
    ):
        super().__init__(**kwargs)
        self.out_min, self.out_max = out_min, out_max
        self.args_names = "out_min", "out_max"

    def apply_transform(self, subject: Subject) -> Subject:
        for image in self.get_images(subject):
            self.apply_threshold(image)
        return subject

    def apply_threshold(self, image: ScalarImage) -> None:
        image.set_data(self.threshold(image.data))

    def threshold(self, tensor: torch.Tensor) -> torch.Tensor:
        C = torch.zeros(tensor.shape, dtype=tensor.dtype)
        l1_tensor = torch.where(tensor < self.out_max, tensor, C)
        output = torch.where(l1_tensor > self.out_min, l1_tensor, C)
        return output


# the "_transform" functions return lambdas that can be used to wrap into a Compose class
def threshold_transform(parameters: dict) -> Threshold:
    """
    This function returns a lambda function that can be used to wrap into a Compose class.

    Args:
        parameters (dict): The parameters dictionary.

    Returns:
        Threshold: The transform to threshold the image.
    """
    return Threshold(out_min=parameters["min"], out_max=parameters["max"])


def clip_transform(parameters: dict) -> Clamp:
    """
    This function returns a lambda function that can be used to wrap into a Compose class.

    Args:
        parameters (dict): The parameters dictionary.

    Returns:
        Clamp: The transform to clip the image.
    """
    return Clamp(out_min=parameters["min"], out_max=parameters["max"])



from pathlib import Path
import tempfile

import numpy as np

from panoptica import Panoptica_Evaluator


def generate_instance_segmentation(
    prediction: np.ndarray,
    target: np.ndarray,
    parameters: dict = None,
    panoptica_config_path: str = None,
) -> dict:
    """
    Evaluate a single exam using Panoptica.

    Args:
        prediction (np.ndarray): The input prediction containing objects.
        label_path (str): The path to the reference label.
        target (np.ndarray): The input target containing objects.
        panoptica_config_path (str): The path to the Panoptica configuration file.

    Returns:
        dict: The evaluation results.
    """

    cwd = Path(__file__).parent.absolute()
    # the parameters dict takes precedence over the panoptica_config_path
    panoptica_config = parameters.get("panoptica_config", None)
    if panoptica_config is None:
        panoptica_config_path = (
            cwd / "panoptica_config_brats.yaml"
            if panoptica_config_path is None
            else panoptica_config_path
        )
    else:
        # write the panoptica config to a file
        panoptica_config_path = tempfile.NamedTemporaryFile(
            mode="w", delete=False, suffix=".yaml"
        ).name
        with open(panoptica_config_path, "w") as f:
            f.write(panoptica_config)
    evaluator = Panoptica_Evaluator.load_from_config(panoptica_config_path)

    # call evaluate
    group2result = evaluator.evaluate(prediction_arr=prediction, reference_arr=target)

    results = {k: r.to_dict() for k, r in group2result.items()}
    return results



import os
from typing import Tuple

import pandas as pd


def handle_collisions(
    df: pd.DataFrame, headers: dict, output_path: str
) -> Tuple[bool, pd.DataFrame]:
    """
    This function checks for collisions in the subject IDs and updates the subject IDs in the dataframe to avoid collisions.

    This function takes a dataframe as input and checks if there are any pairs of subject IDs
    that are similar to each other. If it finds any such pairs, it renames the subject IDs by
    adding a suffix of '_v1', '_v2', or '_v3' to differentiate them. The function then creates
    a new dataframe that can be used for inference purposes. Additionally, it writes the original
    dataframe to disk for future reference and creates a 'collision.csv' file to inform the user
    of any subject ID collisions that were detected during the process.

    Args:
        df (pd.DataFrame): The dataframe containing the subject IDs to be checked for collisions.
        headers (dict): The headers dictionary containing the subjectIDHeader key.
        output_path (str): The path to the output directory where the updated dataframe and the collision.csv file will be written.

    Returns:
        Tuple[bool, pd.DataFrame]: A tuple containing a boolean indicating whether any collisions were found, and the updated dataframe.
    """

    # Find the subjectID header
    subject_id_column_name = headers.get("subjectIDHeader", None)
    assert (
        subject_id_column_name is not None
    ), "No subject ID column found in the headers."

    # Create a dictionary to store the count of each subjectid
    subjectid_counts = {}

    # Create a list to store the colliding subjectids
    collisions = []

    # Create a new dataframe to store the updated subjectids
    new_df = df.copy()

    # Loop through each row in the original dataframe
    for i, row in df.iterrows():
        subjectid = row[subject_id_column_name]

        # If the subjectid has not been seen before, add it to the dictionary
        if subjectid not in subjectid_counts:
            subjectid_counts[subjectid] = 0
        else:
            # If the subjectid has been seen before, increment the count and add it to the collisions list
            subjectid_counts[subjectid] += 1
            collisions.append(subjectid)
            # Update the subjectid in the new dataframe
            new_subjectid = f"{subjectid}_v{subjectid_counts[subjectid]}"
            new_df.at[i, subject_id_column_name] = new_subjectid

    # Write the colliding subject IDs to the collision.csv file
    if collisions:
        collision_path = os.path.join(output_path, "collision.csv")
        pd.DataFrame({subject_id_column_name: collisions}).to_csv(
            collision_path, index=False
        )

    # Write the updated dataframe to the updated_test_mapping.csv file
    mapping_path = os.path.join(output_path, "updated_test_mapping.csv")
    df.to_csv(mapping_path, index=False)

    # Return a tuple indicating whether any collisions were found, and the updated dataframe
    return len(collisions) > 0, df



from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Literal, Union
from GANDLF.schedulers import global_schedulers_dict

TYPE_OPTIONS = Literal[tuple(global_schedulers_dict.keys())]


class BaseTriangleConfig(BaseModel):
    min_lr: float = Field(default=(10**-3))
    max_lr: float = Field(default=1)
    step_size: float = Field(description="step_size", default=None)


class TriangleModifiedConfig(BaseModel):
    min_lr: float = Field(default=0.000001)
    max_lr: float = Field(default=0.001)
    max_lr_multiplier: float = Field(default=1.0)
    step_size: float = Field(description="step_size", default=None)


class CyclicLrBaseConfig(BaseModel):
    # More details https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.CyclicLR.html
    min_lr: float = Field(
        default=None
    )  # The default value is calculated according the learning rate  * 0.001
    max_lr: float = Field(default=None)  # calculate in the validation stage
    gamma: float = Field(default=0.1)
    scale_mode: Literal["cycle", "iterations"] = Field(default="cycle")
    cycle_momentum: bool = Field(default=False)
    base_momentum: float = Field(default=0.8)
    max_momentum: float = Field(default=0.9)
    step_size: float = Field(description="step_size", default=None)


class ExpConfig(BaseModel):
    gamma: float = Field(default=0.1)


class StepConfig(BaseModel):
    gamma: float = Field(default=0.1)
    step_size: float = Field(description="step_size", default=None)


class CosineannealingConfig(BaseModel):
    # More details https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.CosineAnnealingWarmRestarts.html
    T_0: int = Field(default=5)
    T_mult: float = Field(default=1)
    min_lr: float = Field(default=0.001)


class ReduceOnPlateauConfig(BaseModel):
    # More details https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.ReduceLROnPlateau.html
    min_lr: Union[float, list] = Field(default=None)
    gamma: float = Field(default=0.1)
    mode: Literal["min", "max"] = Field(default="min")
    factor: float = Field(default=0.1)
    patience: int = Field(default=10)
    threshold: float = Field(default=0.0001)
    cooldown: int = Field(default=0)
    threshold_mode: Literal["rel", "abs"] = Field(default="rel")


class WarmupcosinescheduleConfig(BaseModel):
    # More details https://docs.monai.io/en/stable/optimizers.html#monai.optimizers.WarmupCosineSchedule
    warmup_steps: int = Field(default=None)


# It allows extra parameters
class SchedulerConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    type: TYPE_OPTIONS = Field(description="scheduler type")


# Define the type and the scheduler base model class
schedulers_dict_config = {
    "triangle": BaseTriangleConfig,
    "triangle_modified": TriangleModifiedConfig,
    "triangular": CyclicLrBaseConfig,
    "exp_range": CyclicLrBaseConfig,
    "exp": ExpConfig,
    "exponential": ExpConfig,
    "step": StepConfig,
    "reduce_on_plateau": ReduceOnPlateauConfig,
    "reduce-on-plateau": ReduceOnPlateauConfig,
    "plateau": ReduceOnPlateauConfig,
    "reduceonplateau": ReduceOnPlateauConfig,
    "cosineannealing": CosineannealingConfig,
    "warmupcosineschedule": WarmupcosinescheduleConfig,
    "wcs": WarmupcosinescheduleConfig,
}
